"""
Created on 6 Jul 2025

@author: ph1jb
"""

from configargparse import Namespace, YAMLConfigFileParser  # type: ignore
from pathlib import Path
from typing import Literal
import configargparse
import dbcache
import dotenv
import logging
import yaml


logger = logging.getLogger(__name__)
type Compression = Literal["bz2", "gzip", "xz", "zlib"]
type Connector = Literal["mysqlconnector", "mysqlclient", "pymysql"]
type Serialisation = Literal["csv", "pickle"]


class Main:
    logformat_default = "%(asctime)s %(module)s %(funcName)s %(lineno)d %(levelname)s %(message)s"
    columns_default = "id,source,source_id,longitude,latitude,install_date,dc_capacity_mwp"

    def __init__(self, config_file: Path, secrets_file: Path):
        """
        Initialise the Main class, parse configuration and set up database connection parameters.
        """
        self.config: Namespace = self._parse_config(config_file, secrets_file)
        self._setup_logging(self.config)
        self.table_source = self.config.table_source
        self.table_cache = self.config.table_cache
        self.sqla_handler = dbcache.Config._create_query_handler(self.config)
        self.cache_manager = dbcache.Main._create_cache_manager(
            self.sqla_handler, self.config.compression, self.config.serialisation
        )

    @staticmethod
    def _parse_config(config_file: Path, secrets_file: Path) -> Namespace:
        """Get parser with specified arguments (CLI, config, env var, default)"""
        p = configargparse.ArgParser(
            config_file_parser_class=YAMLConfigFileParser,
            default_config_files=[config_file, secrets_file],
        )
        p.add(
            "--columns",
            default="*",
            env_var="COLUMNS",
            help="Columns to select from source table",
        )
        p.add(
            "--compression",
            choices=["bz2", "gzip", "xz", "zlib"],
            default="gzip",
            env_var="COMPRESSION",
            help="Compression lib",
        )
        p.add(
            "--connector",
            choices=["mysqlconnector", "mysqlclient", "pymysql"],
            default="pymysql",
            env_var="CONNECTOR",
            help="MySQL connector",
        )
        p.add_argument(
            "--logformat", default=Main.logformat_default, env_var="LOGFORMAT", help="Log format"
        )
        p.add_argument(
            "--loglevel",
            choices=["INFO", "WARNING", "DEBUG", "ERROR", "CRITICAL"],
            default="INFO",
            env_var="LOGLEVEL",
            help="Log level (INFO, WARNING, DEBUG etc)",
        )
        p.add("--mysql_database", env_var="MYSQL_DATABASE", help="Database name")
        p.add("--mysql_host", env_var="MYSQL_HOST", help="Database host")
        p.add("--mysql_user", env_var="MYSQL_USER", help="Database user")
        p.add("--mysql_password", help="Database password.")
        p.add(
            "--mysql_options",
            env_var="MYSQL_OPTIONS",
            required=True,
            type=yaml.safe_load,
            help="MySQL options",
        )
        p.add(
            "--serialisation",
            choices=["csv", "pickle"],
            default="pickle",
            env_var="SERIALISATION",
            help="Database port",
        )
        p.add(
            "--table_source",
            env_var="TABLE_SOURCE",
            required=True,
            help="Source database table name",
        )
        p.add(
            "--table_cache",
            env_var="TABLE_CACHE",
            required=True,
            help="Target database table name",
        )
        logger.debug("Parsing configuration arguments.")
        return p.parse_args()

    @staticmethod
    def _setup_logging(config) -> None:
        """Set up logging."""
        logging.basicConfig(format=config.logformat, level=config.loglevel)

    def run(self):
        self.cache_manager.read_data(
            self.config.table_cache, self.config.table_source, self.config.columns
        )


if __name__ == "__main__":
    dotenv.load_dotenv()
    CONFIG_FILE = Path(__file__).resolve().parent.parent / "config" / "dbcache_config.yml"
    SECRETS_FILE = Path(__file__).resolve().parent.parent / "secrets" / "dbcache_secrets.yml"
    MAIN = Main(CONFIG_FILE, SECRETS_FILE)
    MAIN.run()
